# Basic ns setup
set ns [new Simulator]

set overheadFile [open "results/overhead.txt" a+]
set executionFile [open "results/execution.txt" a+]

proc finish {} {
        #global ns nf
        #$ns flush-trace
        #close $nf
        #exec nam out.nam &
        exit 0
}

proc getdistance { id1 id2 } {
	global routelogic
	set nhops 0

	set nexthop $id1
	
	while {$nexthop != $id2} {
		set nhops [expr $nhops + 1]
		set nexthop [$routelogic lookup $nexthop $id2]
	}

	return $nhops
}

source random.tcl

if {$argc != 7} {
	puts "Wrong syntax!"
	puts "Usage: ns placeAPs.tcl PhysicalNodesNumber VirtualMachinesNumber InformationSourcesNumber InformationSources ManagementApplicationLocation PlacementMethod NumberOfAPs"
	exit 0
}

set PhysicalNodesNumber [lindex $argv 0]
set VirtualMachinesNumber [lindex $argv 1]
set InformationSourcesNumber [lindex $argv 2]
set InformationSources [lindex $argv 3]
set ManagementApplicationLocation [lindex $argv 4]
set PlacementMethod [lindex $argv 5]
set NumberofAPs [lindex $argv 6]

#puts [lindex $VirtualMachinesNumber 0]
#puts [lindex $InformationSources 1]


#VMs connected per Physical Host
for {set i 0} {$i < $PhysicalNodesNumber} {incr i} {
	set number_of_nodes($i) [lindex $VirtualMachinesNumber $i]
	puts "Physical Machine:$i has $number_of_nodes($i) VMs"

}

set AS_number $PhysicalNodesNumber

#transmission rate
set h 1
#resolution rate
set k 10
#message size
set d 100
#number of managed entities
set ME_number $InformationSourcesNumber
#number of managers
#only value 1 works!
set MA_number 1

set bandwidth 1Mbit
set propagation_delay 10ms

#number of KP nodes
set KP_number $NumberofAPs

#kp method:
#1 Random
#2 Hot-spot
#3 Greedy

set KP_method $PlacementMethod

#puts "Placement method: $KP_method"

for {set i 0} {$i < $KP_number} {incr i} {
	set is_KP($i) 0
}

for {set i 0} {$i < $ME_number} {incr i} {
        set ME_location($i) 0
}

set experiment_duration 60.0

set NN -1
set NN_old -1

for {set i 0} {$i < $AS_number} {incr i} {
	set NN [expr $NN + 1]
	set AS($NN) [$ns node]

	if {$NN_old > -1} {
		#puts $NN.$NN_old
		$ns duplex-link $AS($NN_old) $AS($NN) $bandwidth $propagation_delay DropTail
	}

	#if number_of_nodes = -1 then set a random value
	if {$number_of_nodes($i) == -1} {
		set number_of_nodes($i) [get_random_integer 0 10]
	}

	set NN_old $NN

	#each intermediate node should have e1 and e2 nodes connected
	for {set k 0} {$k < $number_of_nodes($i)} {incr k} {
		set NN [expr $NN + 1]
		set AS($NN) [$ns node]

		#puts $NN_old.$NN
		$ns duplex-link $AS($NN_old) $AS($NN) $bandwidth $propagation_delay DropTail
	}

}

#set manager location
for {set i 0} {$i < $MA_number} {incr i} {
	#set current_number [get_random_integer 0 $NN]
	#set MA_location $current_number
	#puts "$current_number is the manager"
	#Management application is green
	set MA_location [lindex $argv 2]
	$AS($MA_location) color green
}


#set managed entities
for {set i 0} {$i < $ME_number} {incr i} {
	#if the random number is not used, proceed
	set used 1
	while {$used} {
		set current_number [get_random_integer 0 $NN]
		set current_number 
		set used 0
		for {set k 0} {$k < $i} {incr k} {
			if {$ME_location($k)==$current_number} {
				set used 1
			}
		}
	}

	#puts $current_number
	set udp($i) [new Agent/UDP]
	$ns attach-agent $AS($current_number) $udp($i)
	set ME_location($i) $current_number
	#puts "$current_number is the client"

	#Management entity is red
	$AS($ME_location($i)) color red

        set cbr($i) [new Application/Traffic/CBR]
        $cbr($i) set packetSize_ $d
        $cbr($i) set interval_ $h
        $cbr($i) attach-agent $udp($i)

	#$ns at 0.0 "$cbr($i) start"
	#$ns at $experiment_duration "$cbr($i) stop"
}

#reset is_KP array
for {set i 0} {$i < $NN} {incr i} {
        set is_KP($i) 0
}


set routelogic [$ns get-routelogic]

$ns compute-routes

proc random_method {} {
	global KP_number
	global NN
	global KP_location
	global is_KP
	global AS

		for {set i 0} {$i < $KP_number} {incr i} {
			#set sinks for KP node
			#if the random number is not used, proceed
			set used 1
			while {$used} {
				set current_number [get_random_integer 0 $NN]
				set used 0
				for {set k 0} {$k < $i} {incr k} {
					if {$KP_location($k)==$current_number} {
						set used 1
					}
				}
			}
			set KP_location($i) $current_number
			set is_KP($current_number) 1 

			#puts "$current_number is a KP node"

			#KP node is yellow
			$AS($current_number) color yellow

		}
}

proc hotspot_method {} {
	global KP_number
	global NN
	global KP_location
	global is_KP
	global AS

		for {set i 0} {$i < $NN} {incr i} {
			set onehopaway($i) 0
			set twohopsaway($i) 0
			set threehopsaway($i) 0

			for {set k 0} {$k < $NN} {incr k} {
				set distance [getdistance $i $k]
				if {$distance == 1} {
					set onehopaway($i) [expr $onehopaway($i) + 1]
				}
				if {$distance == 2} {
					set twohopsaway($i) [expr $twohopsaway($i) + 1]
				}
				if {$distance == 3} {
					set threehopsaway($i) [expr $threehopsaway($i) + 1]
				}

			}
			set cost($i) [expr ($onehopaway($i) * $onehopaway($i) * $onehopaway($i)) + ($twohopsaway($i) * $twohopsaway($i)) + $threehopsaway($i)]
			#puts "Node:$i cost: $cost($i)"
			#puts "Node:$i Onehop:$onehopaway($i) Twohops:$twohopsaway($i) Threehops:$threehopsaway($i)"
		}

		#get the KP_number best options
		
		for {set i 0} {$i < $KP_number} {incr i} {
			set bestcost 0
			for {set k 0} {$k < $NN} {incr k} {
				if {$bestcost < $cost($k)} {
					set bestcost $cost($k)
					set bestnode $k
				}

			}
			#puts "Best cost:$bestcost best node:$bestnode"
			set cost($bestnode) 0
			
			set KP_location($i) $bestnode
			set is_KP($bestnode) 1 

			#puts "$bestnode is a KP node"

			#KP node is yellow
			$AS($bestnode) color yellow
			
		}


}

proc greedy1_method {} {
	global KP_number
	global NN
	global KP_location
	global is_KP
	global AS

		set KP_selected 0
		while {$KP_selected < $KP_number} {
			for {set i 0} {$i < $NN} {incr i} {
				set cost($i) 0
				for {set k 0} {$k < $NN} {incr k} {
					#calculate cost to the nearest KP node, per node
					set best_distance [getdistance $k $i]
					for {set z 0} {$z < $KP_selected} {incr z} {
						#if existing KP node is nearer, select that
						set distance [getdistance $k $KP_location($z)]
						if {$distance < $best_distance} {
							set best_distance $distance
						}
					}
					#puts "node $i best distance:$best_distance"
					set cost($i) [expr $cost($i) + $best_distance]		
				}
				#puts "Cost for $i is $cost($i)"
			}
			#disable existing KP nodes, just in case
			#for {set z 0} {$z < $KP_selected} {incr z} {
			#	set cost($z) 100000
			#}

			#select the node with the best proximity
			set bestcost 100000
			for {set i 0} {$i < $NN} {incr i} {
				if {$bestcost > $cost($i)} {
					set bestcost $cost($i)
					set bestnode $i
				}
			}
			#puts "Best cost: $bestcost best node:$bestnode"
			set KP_location($KP_selected) $bestnode			
			set KP_selected [expr $KP_selected + 1]

			set is_KP($bestnode) 1 

			#KP node is yellow
			$AS($bestnode) color yellow
		}


}

proc greedy2_method {} {
	global KP_number
	global NN
	global KP_location
	global is_KP
	global AS
	global ME_number
	global ME_location

		set KP_selected 0
		while {$KP_selected < $KP_number} {
			for {set i 0} {$i < $NN} {incr i} {
				set cost($i) 0
				for {set k 0} {$k < $ME_number} {incr k} {
					#calculate cost to the nearest KP node, per node
					set best_distance [getdistance $ME_location($k) $i]
					for {set z 0} {$z < $KP_selected} {incr z} {
						#if existing KP node is nearer, select that
						set distance [getdistance $ME_location($k) $KP_location($z)]
						if {$distance < $best_distance} {
							set best_distance $distance
						}
					}
					#puts "node $i best distance:$best_distance"
					set cost($i) [expr $cost($i) + $best_distance]		
				}
				#puts "Cost for $i is $cost($i)"
			}
			#disable existing KP nodes, just in case
			#for {set z 0} {$z < $KP_selected} {incr z} {
			#	set cost($z) 100000
			#}

			#select the node with the best proximity
			set bestcost 100000
			for {set i 0} {$i < $NN} {incr i} {
				if {$bestcost > $cost($i)} {
					set bestcost $cost($i)
					set bestnode $i
				}
			}
			#puts "Best cost: $bestcost best node:$bestnode"
			set KP_location($KP_selected) $bestnode			
			set KP_selected [expr $KP_selected + 1]

			set is_KP($bestnode) 1 

			#KP node is yellow
			$AS($bestnode) color yellow


		}


}


if {$KP_number == 0} {
	#direct connection
	for {set i 0} {$i < $ME_number} {incr i} {
		for {set k 0} {$k < $MA_number} {incr k} {
			set null($i) [new Agent/LossMonitor]
			$ns attach-agent $AS($MA_location) $null($i)

                	$ns connect $udp($i) $null($i)
                	#puts "connect $i $k"
        	}
	}
} else {
	#place KP nodes
	set placementtime 0
	if {$KP_method == 1} {
		#random method
		
		#exec random placement and calculate execution time
		set placementtime [time {random_method}]
		puts $placementtime

	}
	if {$KP_method == 2} {
		#hotspot method
		
		#exec hotspot placement and calculate execution time
		set placementtime [time {hotspot_method}]
		puts $placementtime

	}

	if {$KP_method == 3} {
		#greedy method

		#exec greedy1 placement and calculate execution time
		set placementtime [time {greedy1_method}]
		puts $placementtime


	}

	if {$KP_method == 4} {
		#greedy2 method

		#exec greedy2 placement and calculate execution time
		set placementtime [time {greedy2_method}]
		puts $placementtime

	}

	puts $executionFile $placementtime


	#communicate through the nearest KP node
	for {set k 0} {$k < $KP_number} {incr k} {
			set MAKP_distance($k) [getdistance $MA_location $KP_location($k)]
			#puts "MAKP_distance:$k $MAKP_distance($k)"
	}


	for {set i 0} {$i < $ME_number} {incr i} {
		set best_distance 100
		
		for {set k 0} {$k < $KP_number} {incr k} {
			set KP_distance($i,$k) [getdistance $ME_location($i) $KP_location($k)]
			if {$best_distance > $KP_distance($i,$k)} {
				set best_distance $KP_distance($i,$k)
				set best_KP_location($i) $KP_location($k)
				set best_KP($i) $k

			}

			if {$best_distance == $KP_distance($i,$k)} {
				#if they have equal distance, get the KP that is closer to the MA
				if {$MAKP_distance($k) < $MAKP_distance($best_KP($i))} {
					set best_distance $KP_distance($i,$k)
					set best_KP_location($i) $KP_location($k)
					set best_KP($i) $k
				}				
			}

		}
		#puts "MEKP_distance: $ME_location($i).$KP_location($k).$KP_distance($i,$k)"
		#puts "Best distance:$best_distance Best KP:$best_KP($i)"
	}

	for {set i 0} {$i < $ME_number} {incr i} {
		#puts "Best KP for ME:$ME_location($i) is $best_KP_location($i)"
	}

	#connect ME with KP nodes
	for {set i 0} {$i < $ME_number} {incr i} {
		set nullKP($i) [new Agent/LossMonitor]
		$ns attach-agent $AS($best_KP_location($i)) $nullKP($i)

               	$ns connect $udp($i) $nullKP($i)
	}

	#connect KP nodes with MA
	for {set i 0} {$i < $ME_number} {incr i} {

		set null($i) [new Agent/LossMonitor]
		$ns attach-agent $AS($MA_location) $null($i)


		set udpKP($i) [new Agent/UDP]
		$ns attach-agent $AS($best_KP_location($i)) $udpKP($i)
        	set cbrKP($i) [new Application/Traffic/CBR]
        	$cbrKP($i) set packetSize_ $d
        	$cbrKP($i) set interval_ $k
        	$cbrKP($i) attach-agent $udpKP($i)

		$ns at 0.0 "$cbrKP($i) start"
		$ns at $experiment_duration "$cbrKP($i) stop"
		
		$ns connect $udpKP($i) $null($i)
		#puts "connect KP:$best_KP_location($i) with MA"
	}
} 

#calculate overhead
set totaloverhead 0
set totaldistance 0
for {set i 0} {$i < $ME_number} {incr i} {
	#if KP are not used
	if {$KP_number == 0} {
		set distance [getdistance $ME_location($i) $MA_location]
		set overhead [expr $distance * $h * $d]	
		
		set totaloverhead [expr $totaloverhead + $overhead]
		set totaldistance [expr $totaldistance + $distance]
	
	} else {
		set MEKPdistance [getdistance $ME_location($i) $best_KP_location($i)]
		set MEKPoverhead [expr 1.0 * $MEKPdistance * ($d / $h)]

		set KPMAdistance [getdistance $best_KP_location($i) $MA_location]
		set KPMAoverhead [expr 1.0 * $KPMAdistance * ($d / $k)]

		set totaloverhead [expr $totaloverhead + $MEKPoverhead + $KPMAoverhead]
		set totaldistance [expr $totaldistance + $MEKPdistance + $KPMAdistance]
		
	}
}
set averageoverhead [expr 1.0 * ($totaloverhead / $totaldistance)] 

puts $averageoverhead

puts $overheadFile $averageoverhead

$ns run
$ns at $experiment_duration "finish"

#set totalbytes [$null(0) set bytes_]

#puts $totalbytes
close $executionFile
close $overheadFile
